@@ -200,6 +200,15 @@ static int _pitenumdel(_hitem *item, void *arg);
 
 // funcs
 
+static PyCodeObject *
+FRAME2CODE(PyFrameObject *frame) {
+#if PY_MAJOR_VERSION >= 3 && PY_MINOR_VERSION >= 10
+    return PyFrame_GetCode(frame);
+#else
+    return frame->f_code;
+#endif
+}
+
 static void _DebugPrintObjects(unsigned int arg_count, ...)
 {
     unsigned int i;
@@ -214,9 +223,18 @@ static void _DebugPrintObjects(unsigned int arg_count, ...)
 }
 
 int 
-IS_SUSPENDED(PyFrameObject *frame)
-{
-#if PY_MAJOR_VERSION >= 3 && PY_MINOR_VERSION >= 10
+IS_SUSPENDED(PyFrameObject *frame) {
+#if PY_VERSION_HEX >= 0x030B0000  // Python 3.11+
+    PyGenObject *gen = (PyGenObject *)PyFrame_GetGenerator(frame);
+    if (gen == NULL) {
+        return 0;
+    }
+
+    // -1 is FRAME_SUSPENDED. See internal/pycore_frame.h
+    // TODO: Remove these after 3.12 make necessary public APIs.
+    // See https://discuss.python.org/t/python-3-11-frame-structure-and-various-changes/17895
+    return gen->gi_frame_state == -1;
+#elif PY_MAJOR_VERSION >= 3 && PY_MINOR_VERSION == 10
     return (frame->f_state == FRAME_SUSPENDED);
 #else
     return (frame->f_stacktop != NULL);
@@ -229,11 +247,11 @@ int IS_ASYNC(PyFrameObject *frame)
 
 #if defined(IS_PY3K) 
 #if PY_MINOR_VERSION >= 4
-    result = frame->f_code->co_flags & CO_COROUTINE || 
-        frame->f_code->co_flags & CO_ITERABLE_COROUTINE;
+    result = FRAME2CODE(frame)->co_flags & CO_COROUTINE || 
+        FRAME2CODE(frame)->co_flags & CO_ITERABLE_COROUTINE;
 #endif
 #if PY_MINOR_VERSION >= 6
-    result = result || frame->f_code->co_flags & CO_ASYNC_GENERATOR;
+    result = result || FRAME2CODE(frame)->co_flags & CO_ASYNC_GENERATOR;
 #endif
 #endif
 
@@ -625,6 +643,17 @@ _ccode2pit(void *cco, uintptr_t current_tag)
     return ((_pit *)it->val);
 }
 
+static PyObject *_get_locals(PyFrameObject *fobj) {
+#if PY_MAJOR_VERSION >= 3 && PY_MINOR_VERSION >= 10
+    return PyEval_GetLocals();
+#else
+    PyFrame_FastToLocals(fobj);
+    PyObject* locals = fobj->f_locals;
+    PyFrame_LocalsToFast(fobj, 0);
+    return locals;
+#endif
+}
+
 // maps the PyCodeObject to our internal pit item via hash table.
 static _pit *
 _code2pit(PyFrameObject *fobj, uintptr_t current_tag)
@@ -633,13 +662,14 @@ _code2pit(PyFrameObject *fobj, uintptr_t current_tag)
     PyCodeObject *cobj;
     _pit *pit;
     _htab *pits;
+    PyObject *co_varnames;
 
     pits = _get_pits_tbl(current_tag);
     if (!pits) {
         return NULL;
     }
 
-    cobj = fobj->f_code;
+    cobj = FRAME2CODE(fobj);
     it = hfind(pits, (uintptr_t)cobj);
     if (it) {
         return ((_pit *)it->val);
@@ -658,12 +688,18 @@ _code2pit(PyFrameObject *fobj, uintptr_t current_tag)
     pit->fn_descriptor = (PyObject *)cobj;
     Py_INCREF(cobj);
 
-    PyFrame_FastToLocals(fobj);
     if (cobj->co_argcount) {
-        const char *firstarg = PyStr_AS_CSTRING(PyTuple_GET_ITEM(cobj->co_varnames, 0));
+        // There has been a lot going on with `co_varnames`, but finally in 
+        // 3.11.0rc1, it is added as a public API
+#if PY_VERSION_HEX >= 0x030B0000 // Python 3.11+
+        co_varnames = PyCode_GetVarnames(cobj);
+#else
+        co_varnames = cobj->co_varnames;
+#endif
+        const char *firstarg = PyStr_AS_CSTRING(PyTuple_GET_ITEM(co_varnames, 0));
 
         if (!strcmp(firstarg, "self")) {
-            PyObject* locals = fobj->f_locals;
+            PyObject* locals = _get_locals(fobj);
             if (locals) {
                 PyObject* self = PyDict_GetItemString(locals, "self");
                 if (self) {
@@ -685,8 +721,6 @@ _code2pit(PyFrameObject *fobj, uintptr_t current_tag)
         pit->name = cobj->co_name;
     }
 
-    PyFrame_LocalsToFast(fobj, 0);
-
     return pit;
 }
 
@@ -859,6 +893,18 @@ _get_frame_elapsed(void)
     return result;
 }
 
+static void
+_print_coros(_pit *cp)
+{
+    _coro *coro;
+
+    printf("Printing coroutines on %s...\n", PyStr_AS_CSTRING(cp->name));
+    coro = cp->coroutines;
+    while(coro) {
+        printf("CORO %s %p %lld\n", PyStr_AS_CSTRING(cp->name), coro->frame, coro->t0);
+        coro = (_coro *)coro->next;
+    }
+}
 
 static int 
 _coro_enter(_pit *cp, PyFrameObject *frame)
@@ -879,7 +925,7 @@ _coro_enter(_pit *cp, PyFrameObject *frame)
         coro = (_coro *)coro->next;
     }
 
-    //printf("CORO ENTER %s %p\n", PyStr_AS_CSTRING(cp->name), frame);
+    //printf("CORO ENTER %s %p %lld\n", PyStr_AS_CSTRING(cp->name), frame, tickcount());
 
     coro = ymalloc(sizeof(_coro));
     if (!coro) {
@@ -909,7 +955,7 @@ _coro_exit(_pit *cp, PyFrameObject *frame)
             return 0;
     }
 
-    //printf("CORO EXIT %s %p\n", PyStr_AS_CSTRING(cp->name), frame);
+    //printf("CORO EXIT %s %p %lld\n", PyStr_AS_CSTRING(cp->name), frame, tickcount());
 
     prev = NULL;
     coro = cp->coroutines;
@@ -919,9 +965,10 @@ _coro_exit(_pit *cp, PyFrameObject *frame)
             if (prev) {
                 prev->next = coro->next;
             } else {
-                cp->coroutines = NULL;
+                cp->coroutines = (_coro *)coro->next;
             }
             yfree(coro);
+            //printf("CORO EXIT(elapsed) %s %p %lld\n", PyStr_AS_CSTRING(cp->name), frame, tickcount()-_t0);
             return tickcount() - _t0;
         }
         prev = coro;
@@ -1010,7 +1057,7 @@ _call_leave(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
     }
 
     // if the function that the frame belongs is a coroutine, we check if we RETURN
-    // or YIELD from the coroutine to calculate the correct walltime
+    // or await the coroutine to calculate the correct walltime
     if (IS_ASYNC(frame)) {
         if (IS_SUSPENDED(frame)) {
             yielded = 1;
@@ -1255,11 +1302,38 @@ _resume_greenlet_ctx(_ctx *ctx)
     ydprintf("resuming context: %ld, shift: %lld", ctx->id, shift);
 }
 
+static void 
+_eval_setprofile(PyThreadState *ts)
+{
+#if PY_VERSION_HEX > 0x030b0000
+    _PyEval_SetProfile(ts, _yapp_callback, NULL);
+#elif PY_VERSION_HEX < 0x030a00b1
+    ts->use_tracing = 1;
+    ts->c_profilefunc = _yapp_callback;
+#else
+    ts->cframe->use_tracing = 1;
+    ts->c_profilefunc = _yapp_callback;
+#endif
+}
+
+static void
+_eval_unsetprofile(PyThreadState *ts)
+{
+#if PY_VERSION_HEX > 0x030b0000
+    _PyEval_SetProfile(ts, NULL, NULL);
+#elif PY_VERSION_HEX < 0x030a00b1
+    ts->use_tracing = 0;
+    ts->c_profilefunc = NULL;
+#else
+    ts->cframe->use_tracing = 0;
+    ts->c_profilefunc = NULL;
+#endif
+}
+
 static _ctx *
 _bootstrap_thread(PyThreadState *ts)
 {
-    ts->use_tracing = 1;
-    ts->c_profilefunc = _yapp_callback;
+    _eval_setprofile(ts);
     return NULL;
 }
 
@@ -1288,9 +1362,7 @@ _profile_thread(PyThreadState *ts)
     } else {
         ctx = (_ctx *)it->val;
     }
-    
-    ts->use_tracing = 1;
-    ts->c_profilefunc = _yapp_callback;
+    _eval_setprofile(ts);
     ctx->id = ctx_id;
     ctx->tid = ts->thread_id;
     ctx->ts_ptr = ts;
@@ -1306,8 +1378,7 @@ _profile_thread(PyThreadState *ts)
 static _ctx*
 _unprofile_thread(PyThreadState *ts)
 {
-    ts->use_tracing = 0;
-    ts->c_profilefunc = NULL;
+    _eval_unsetprofile(ts);
 
     return NULL; //dummy return for enum_threads() func. prototype
 }
@@ -1604,6 +1675,7 @@ _pitenumstat(_hitem *item, void *arg)
     _pit *pt;
     PyObject *exc;
     PyObject *children;
+    PyObject *ctx_name;
     _pit_children_info *pci;
     _ctxfuncenumarg *eargs;
 
@@ -1644,11 +1716,19 @@ _pitenumstat(_hitem *item, void *arg)
     if (pt->callcount == 0)
         pt->callcount = 1;
 
+    // normally _yapp_callback sets ctx->name if not set but there is a possibility
+    // that we might end up here even it is not set. We also do not default it to 
+    // Py_None as we will Py_CLEAR() on it later on.
+    ctx_name = eargs->ctx->name;
+    if (!ctx_name) {
+        ctx_name = Py_None;
+    }
+
     exc = PyObject_CallFunction(eargs->enum_args->enumfn, "((OOkkkIffIOkOkO))", 
                         pt->name, pt->modname, pt->lineno, pt->callcount,
                         pt->nonrecursive_callcount, pt->builtin, 
                         _normt(pt->ttotal), _normt(pt->tsubtotal),
-                        pt->index, children, eargs->ctx->id, eargs->ctx->name,
+                        pt->index, children, eargs->ctx->id, ctx_name,
                         eargs->tag, pt->fn_descriptor);
 
     if (!exc) {
@@ -1727,7 +1807,7 @@ start(PyObject *self, PyObject *args)
 }
 
 static PyObject*
-stop(PyObject *self)
+stop(PyObject *self, PyObject *args)
 {
     _stop();
     Py_RETURN_NONE;
@@ -2081,13 +2161,13 @@ _resume(PyObject *self, PyObject *args)
             // error
             return NULL;
     }
-    
+
     Py_RETURN_NONE;
 }
 
 static PyMethodDef yappi_methods[] = {
     {"start", start, METH_VARARGS, NULL},
-    {"stop", (PyCFunction)stop, METH_NOARGS, NULL},
+    {"stop", stop, METH_NOARGS, NULL},
     {"enum_func_stats", enum_func_stats, METH_VARARGS, NULL},
     {"enum_context_stats", enum_context_stats, METH_VARARGS, NULL},
     {"enum_thread_stats", enum_context_stats, METH_VARARGS, NULL},
@@ -2108,7 +2188,7 @@ static PyMethodDef yappi_methods[] = {
     {"_profile_event", profile_event, METH_VARARGS, NULL}, // for internal usage.
     {"_pause", _pause, METH_VARARGS, NULL}, // for internal usage.
     {"_resume", _resume, METH_VARARGS, NULL}, // for internal usage.
-    {NULL, NULL}      /* sentinel */
+    {NULL, NULL, 0, NULL}      /* sentinel */
 };
 
 #ifdef IS_PY3K
@@ -2158,6 +2238,7 @@ init_yappi(void)
     test_timings = NULL;
 
     SUPPRESS_WARNING(_DebugPrintObjects);
+    SUPPRESS_WARNING(_print_coros);
 
     if (!_init_profiler()) {
         PyErr_SetString(YappiProfileError, "profiler cannot be initialized.");
